def format_code(code_string, indents=1):
    raw_lines = code_string.strip().split("\n")
    indent_list = [len(line) - len(line.lstrip(" ")) for line in raw_lines]
    # Get the first non-zero indent
    base_indent = min(set(indent_list) - set([0]))
    stripped_lines = [line.replace(" " * base_indent, "") for line in raw_lines]
    indent_str = " " * 4 * indents
    return indent_str + f"\n{indent_str}".join(stripped_lines)



trends_page = {
    "dataid": "covid",
    "banks": {"axis": {"type": "axis_controls"}, "trend": {"type": "graph"},},
    "layout": [["axis"], ["trend"]],
    "connections": {"axis": {"trend"}},
}

descriptor = {
    "appbar": {
        "title": "Sample COVID data",
        "subtitle": "A simple, example Bento dashboard",
    },
    "data": {"covid": {"module": "bento.examples.covid"}},
    "pages": {"trends": trends_page},
}


if __name__ == "__main__":
    from bento import bento

    bento.Bento(descriptor).write()



from bento.common import logger, util

logging = logger.fancy_logger(__name__)


def load():
    filename = "sample_stock_data.csv"
    data = {
        "df": util.df_loader(filename, parse_dates=False),
        "keys": ["symbol"],
        "types": {
            "date": "date",
            "symbol": str,
            "open": float,
            "low": float,
            "high": float,
            "close": float,
            "volume": int,
        },
    }
    return data



# These are CSS styles that roughly implement Material Design components
from bento.common import dictutil
from bento import named_themes

import colorsys


def elevation(elev):
    if not elev:
        return {}
    return {"boxShadow": f"1px 1px {elev}px 1px #888888"}


def grid_columns(n_col, pad_pct):
    col_size = (100 + pad_pct) / n_col - pad_pct
    return f"repeat({n_col}, {col_size:.2f}%)"


def flip_color(hex_color, offset_factor=1):
    """ Mirrors the hex color along the brightness axis, e.g #eeeeee -> #111111 """
    if len(hex_color) != 7:
        raise Exception(f"Color {hex_color} is not in #rrggbb format.")
    rgb = [int(hex_color[x : x + 2], 16) for x in [1, 3, 5]]
    hls = list(colorsys.rgb_to_hls(*rgb))
    # Reverse the lightness level
    hls[1] = 255 - hls[1]
    new_rgb = colorsys.hls_to_rgb(*hls)
    rgb2 = [min(255, max(0, int(level))) for level in new_rgb]
    return f"#{rgb2[0]:02x}{rgb2[1]:02x}{rgb2[2]:02x}"


def darken_color(hex_color, offset_factor=0.7):
    """ Mirrors the hex color along the brightness axis, e.g #eeeeee -> #111111 """
    if len(hex_color) != 7:
        raise Exception(f"Color {hex_color} is not in #rrggbb format.")
    rgb = [int(hex_color[x : x + 2], 16) for x in [1, 3, 5]]
    hls = list(colorsys.rgb_to_hls(*rgb))
    # Reduce lightness as defined by offset_factor
    hls[1] *= offset_factor
    new_rgb = colorsys.hls_to_rgb(*hls)
    rgb2 = [min(255, max(0, int(level))) for level in new_rgb]
    return f"#{rgb2[0]:02x}{rgb2[1]:02x}{rgb2[2]:02x}"


theme_keywords = {
    "tight": {"pad_unit": 2, "pad_pct": 0.5, "font__size": "1.3rem"},
    "sparse": {"pad_unit": 8, "pad_pct": 1.5, "font__size": "1.7rem"},
    "loose": {"pad_unit": 16, "pad_pct": 2.5, "font__size": "1.9rem"},
    "flat": {"elevation": 0,},
}


class BentoStyle:
    def __init__(self, theme="", theme_dict=None, layout=(12, 12)):
        # Default to the Bento theme
        self.spec = named_themes.bento
        self.spec["class_name"] = "bento-theme"  # Required for dropdowns, etc.

        # Passing a string assumes using theme keywords e.g. "dark flat"
        for word in theme.split(" "):
            self.spec.update(theme_keywords.get(word, {}))
            if word == "dark":
                for key in self.spec:
                    if "color" in key:
                        if "primary" in key or "secondary" in key:
                            self.spec[key] = darken_color(self.spec[key])
                        else:
                            self.spec[key] = flip_color(self.spec[key])
                self.spec["map__mapbox_style"] = "carto-darkmatter"
        if theme_dict:
            self.spec.update(dictutil.flatten(theme_dict))

        self.layout = {"row": layout[0], "columns": layout[1]}

        # Main defines the whole-page theme which will get inherited by children
        self.main = {
            "height": "100vmax",
            "textAlign": "center",
            "fontFamily": self.spec["font__family"],
            "fontSize": self.spec["font__size"],
            "color": self.spec["color__on_surface"],
            "backgroundColor": self.spec["color__background"],
        }

        self.page = {}
        self.appbar = {
            **elevation(self.spec["elevation"]),
            "backgroundColor": self.spec["color__primary"],
            "color": self.spec["color__on_primary"],
            "display": "flex",
            "flexDirection": "row",
            "justifyContent": "space-between",
            "background-size": "100%",
        }

        # Most components besdies the App Bar are contained in the page
        self.titles = {
            "display": "flex",
            "flexDirection": "row",
            "margin": 0,
            "paddingLeft": 24,
            "alignSelf": "flex-start",
            "alignItems": "flex-end",
        }
        self.h1 = {"margin": 0}
        self.h3 = {"margin": 0, "paddingLeft": 24, "fontStyle": "italic"}
        self.link_set = {"display": "flex", "flexDirection": "row"}
        self.link = {"display": "flex"}
        self.button = {"align-self": "center"}

        self.grid = {
            "display": "grid",
            "paddingTop": f"{self.spec['pad_pct']}%",
            "gridGap": f"{self.spec['pad_pct']}%",
            "rowGap": f"{self.spec['pad_pct'] * 1.4}%",
            "gridTemplateColumns": grid_columns(
                self.layout["columns"], self.spec["pad_pct"]
            ),
            "gridTemplateRows": "auto",
        }

        # Special components
        self.graph = {
            # TODO Figure out how to add this in as additional margin?
            # "margin": {
            #     "l": 16 + 4 * self.spec["pad_unit"],
            #     "b": 40 + 4 * self.spec["pad_unit"],
            #     "t": 40 + 4 * self.spec["pad_unit"],
            #     "r": 40 + 4 * self.spec["pad_unit"],
            # },
            # "transition": {"duration": 500},
            "font": {"color": self.spec["color__on_surface_secondary"]},
            "yaxis": {"gridcolor": self.spec["color__primary"]},
            "xaxis": {"gridcolor": self.spec["color__primary"]},
            "autosize": True,
            "hovermode": "closest",
            "paper_bgcolor": self.spec["color__surface"],
            "plot_bgcolor": self.spec["color__plot_bg"],
            # For maps
            "mapbox_style": self.spec["map__mapbox_style"],
            "geo_oceancolor": "LightBlue",
            "geo_lakecolor": "Blue",
            "geo_landcolor": self.spec["color__plot_bg"],
            "geo_bgcolor": self.spec["color__surface"],
        }

        self.trace = {}

        # Smaller scale objects: Paper > Bar > Block
        self.paper = {
            **elevation(self.spec["elevation"]),
            "backgroundColor": self.spec["color__surface"],
            # "textAlign": "center",
            "display": "flex",
            "flexDirection": "column",
            "padding": self.spec["pad_unit"],
        }

        self.bar = {
            "display": "flex",
            "flex-direction": "row",
            "justify-content": "space-evenly",
        }

        self.label = {}

        self.block = {
            "display": "flex",
            "flexDirection": "column",
            "flexGrow": "1",
            "padding": self.spec["pad_unit"],
        }

        # NOTE We need an entry here for every component library, as is
        self.html = {}
        self.dcc = {}
        # NOTE This fixes x-scrolling for data_table
        self.dash_table = {"overflowX": "hidden"}



import logging
import os
import sys

# Use prettyprinter for dataclass formatting
try:
    import prettyprinter as pprint

    # NOTE ipython extra has issue, throwing a warning if installed and log a DF
    # 'ZMQInteractiveShell' object has no attribute 'highlighting_style'
    pprint.install_extras(exclude=["django", "ipython"], warn_on_error=False)

    # Set a fairly aggressive limit to sequence by default
    # This makes printing more similar to pandas truncation, which is nice
    # NOTE However, this doesn't seem to work for big nested structures, so we
    # might still have to apply a limit on total formatted string lines
    # TODO Add a way to bypass sequence length when needed
    pprint.set_default_config(
        style="dark", max_seq_len=20, width=79, ribbon_width=71, depth=None,
    )
except ImportError:
    import pprint

# ANSI codes that will generate colored text
def_color_map = {
    # reset should suffix any use of the other following prefixes
    "reset": "\x1b[0m",
    "brightred": "\x1b[31m",
    "red": "\x1b[31m",
    "green": "\x1b[32m",
    "yellow": "\x1b[33m",
    "blue": "\x1b[34m",
    "magenta": "\x1b[35m",
    "cyan": "\x1b[36m",
    "light_blue": "\x1b[94m",
    "darkgreen": "\x1b[92m",
    "purple": "\x1b[95m",
}

def_level_color = {
    "CRITICAL": "brightred",
    "ERROR": "red",
    "WARNING": "yellow",
    "INFO": "darkgreen",
    "DEBUG": None,
}

# We never truly want infinite output, 1000 lines is probably enough
config = {
    "DEBUG": {"max_lines": 1000},
    "INFO": {"max_lines": 10},
    "default": {"max_lines": 50},
}

formats = {
    "msg": "{msg}",
    "bare": "{level_str}| {msg}",
    "simple": "{level_str}| {name} | {msg}",
    "full": "{level_str}| {name} | {funcName}:{lineno} - {msg}",
}


# The default schema uses simpler logging for info-level logs, as these are
# intended for consumption at all times (non-debugging)
schemas = {
    "default": {"INFO": "simple", "default": "full"},
}


class FancyFormatter(logging.Formatter):
    """Adds colors and structure to a log output"""

    def __init__(self, schema="default", fmt=None, level_color=None):
        super().__init__()

        # The schema will have level-dependent format strings
        self.schema = schemas.get(schema, schemas["default"])

        # Overrides a schema with a constant log message format
        self.fmt = fmt

        # Level colors can also be overridden, if desired
        self.level_color = level_color or def_level_color
        self.color_map = def_color_map

    def color_text(self, text, color=None):
        """Wraps text in a color code"""
        if not color:
            return text
        prefix = self.color_map.get(color, "")
        if not prefix:
            logging.warning(f"{color} not defined in color_map")
        suffix = self.color_map["reset"] if prefix else ""
        return f"{prefix}{text}{suffix}"

    def level_fmt(self, level):
        return self.color_text(f"{level: <8}", self.level_color.get(level))

    def format(self, record):
        """Automatically called when logging a record"""
        # Allows modification of the record attributes
        record_dict = vars(record)

        # Checks for message format in order:  from message, self.fmt, self.schema
        style = record_dict.get("fmt") or self.fmt
        if not style:
            style = self.schema.get(record.levelname, self.schema["default"])

        # Prepare a pretty version of the message
        curr_conf = config.get(record.levelname, config["default"])
        pretty = pprint.pformat(record_dict["msg"]).strip("'\"")
        total_lines = pretty.count("\n")
        if total_lines > curr_conf["max_lines"]:
            lines = pretty.splitlines()
            # TODO Abstract away the lines left after truncation (e.g. the 2's and 4)
            trunc = total_lines - 6
            pretty = "\n".join(lines[:3] + [f"...truncated {trunc} lines"] + lines[-3:])
        record_dict["level_str"] = self.level_fmt(record.levelname)
        record_dict["msg"] = pretty

        # --- Logging macros for doing some special operations ---
        # Uses purple to draw extra attention to the text of the line
        if record_dict["msg"].startswith("#!"):
            record_dict["msg"] = (
                self.color_text(record_dict["msg"][2:], "purple") + "\n"
            )
        # Doesn't add a newline, allowing line continuation
        elif record_dict["msg"].startswith("#^"):
            record_dict["msg"] = record_dict["msg"][2:]
        # Next two Continue previous line by using a logformat with message text only
        # First with green, indicating a desirable result
        elif record_dict["msg"].startswith("#$+"):
            record_dict["msg"] = self.color_text(record_dict["msg"][3:], "green") + "\n"
            style = "msg"
        # Seconde with red, indicating a problem
        elif record_dict["msg"].startswith("#$-"):
            record_dict["msg"] = self.color_text(record_dict["msg"][3:], "red") + "\n"
            style = "msg"
        else:
            record_dict["msg"] += "\n"
        message = formats[style].format(**record_dict)
        return message


def fancy_logger(name, fmt=None, level=None):
    env_level = os.environ.get("LOGLEVEL_NAME") or os.environ.get("LOGLEVEL")
    if not level:
        level = int(env_level) if env_level else logging.INFO
    name_logger = logging.getLogger(name)
    name_logger.setLevel(level)
    name_logger.propagate = False
    # Make sure not to re-add the handlers if the same name is used
    if not name_logger.handlers:
        handler = logging.StreamHandler(sys.stdout)
        handler.terminator = ""
        handler.setFormatter(FancyFormatter(fmt=fmt))
        name_logger.addHandler(handler)
    return name_logger


if __name__ == "__main__":
    nl = fancy_logger("test")
    for level in ["debug", "info", "warning", "error"]:
        getattr(nl, level)(f"This is a {level} test")
    nl.info("%Purple message")
    nl.info(
        {
            "title": "This is a pretty dictionary",
            "reasons": ["indentation", "length-checking"],
            "strings": [str(i) * 8 for i in range(8)],
            # "codes": [str(i) * 8 for i in range(50)],
        }
    )

    try:
        from dataclasses import dataclass, field
        from typing import List, Dict

        @dataclass
        class Build:
            command: str
            orders: List[str] = field(default_factory=list)
            request: Dict = field(default_factory=lambda: {"url": "gen_build"})

        orders = [
            "Dataclass prettyprint",
            "APT git curl wget",
            "APT build-essential",
            "PIP3 numpy",
            "PIP3 flask GitPython gcp resumable-media",
        ]

        nl.warning(Build(command="1", orders=orders))
    except ImportError:
        pass



from bento.common import logger, util

logging = logger.fancy_logger(__name__)


def load():
    filename = "sample_covid_data.csv"
    data = {
        "df": util.df_loader(filename),
        "keys": ["county", "state", "fips"],
        "types": {"date": "date", "cases": int, "deaths": int,},
    }
    return data



# import dash
import numpy as np
import pandas as pd

from bento.common import logger, logutil, codeutil, dictutil  # noqa
import bento.components as bc

from bento import util as butil


logging = logger.fancy_logger(__name__)


class BentoBanks:
    def __init__(self, data):
        self.data = data
        # We record all registered IDs to help with connections later

        self.outputs = {}
        self.callbacks = {}

        self.connectors = {}

    # @logutil.loginfo(level='debug')
    def _align(self, blocks, vertical, block_size):
        if vertical:
            layout = np.vstack(blocks)
            ideal_height = int(len(blocks) * block_size["ideal"][0])
            min_height = int(len(blocks) * block_size["min"][0])
            sizing = {
                "ideal": [ideal_height, block_size["ideal"][1]],
                "min": [min_height, block_size["min"][1]],
            }
        else:
            layout = np.hstack(blocks)
            ideal_width = int(len(blocks) * block_size["ideal"][1])
            min_width = int(len(blocks) * block_size["min"][1])
            sizing = {
                "ideal": [block_size["ideal"][0], ideal_width],
                "min": [block_size["min"][0], min_width],
            }

        return layout, sizing

    # @logutil.loginfo(level='debug')
    def axis_controls(
        self, gid, dataid, use=("x", "y"), scale=True, vertical=False, **kwargs
    ):
        blocks = []
        option_args = dictutil.extract(r"^options$|^default$", kwargs)
        for axis_idx, axis in enumerate(use):
            # Define the dropdown to select a column to display on the axis
            label = f"{axis}-Axis Data".title()
            id_dict = {"name": f"{axis}_column", **gid}
            cargs = {"Dropdown.clearable": False, **kwargs}

            # NOTE Toggle these lines for early multi-axis support
            # if axis != "x":
            #     cargs["Dropdown.multi"] = True

            default_idx = min(axis_idx, len(self.data[dataid]["columns"]) - 1)
            options = {
                "options": self.data[dataid]["columns"],
                "default": self.data[dataid]["columns"][default_idx],
                **option_args,
            }
            drop_id, dropdown = bc.dropdown(id_dict, options, label=label, **cargs)
            self.outputs[drop_id] = "value"

            if not scale:
                blocks.append([[dropdown]])
                continue

            # Define the radio buttons to select the axis scale
            id_dict = {"name": f"{axis}_scale", **gid}
            radio_id, radio = bc.radio(id_dict, ["linear", "log"])
            self.outputs[radio_id] = "value"

            # The radio buttons will be fed by the selected axis
            self.connectors[radio_id] = {
                "inputs": [(drop_id, "value")],
                "outputs": [(radio_id, "options"), (radio_id, "value")],
            }

            # TODO includes temporary multi-column support
            callback_name = f"{gid['pageid']}_{gid['bankid']}__update_{axis}_radio"
            callback_code = f"""
                inputs = dictutil.process_inputs(dash.callback_context.inputs)
                column = dictutil.extract_unique("_column", inputs)

                if isinstance(column, list):
                    column = column[0]

                col_type = data["{dataid}"]['types'].get(column, 'all')
                options = ["linear", "log", "date"]
                if col_type in (float, int):
                    options = ['linear', 'log']
                elif col_type in ('date', 'datetime'):
                    options = ['date']
                option_dict = butil.gen_options(options)
                return option_dict["options"], option_dict["value"]
                """

            self.callbacks[radio_id] = {
                "provides": ["options", "values"],
                "name": callback_name,
                "code": codeutil.format_code(callback_code),
            }

            blocks.append([[dropdown], [radio]])

        block_size = {"ideal": [1, 2], "min": [1, 1]}
        return self._align(blocks, vertical, block_size)

    # TODO In Progress
    def analytics_set(
        self, gid, dataid, normalize=True, calculus=True, vertical=False, **kwargs
    ):
        blocks = []
        if normalize:
            id_dict = {"name": f"norm_transform", **gid}
            label = f"Normalize by:".title()
            options = {
                "options": ["None", "Max", "Other Series"],
                "default": "None",
            }
            drop_id, dropdown = bc.dropdown(id_dict, options, label=label, **kwargs)
            self.outputs[drop_id] = "value"
            blocks.append([[dropdown]])

        if calculus:
            id_dict = {"name": f"calc_transform", **gid}
            label = f"Calculus (differential, cumulative):".title()
            options = {
                "options": ["Acceleration", "Diff", "None", "Cumulative"],
                "default": "None",
            }
            drop_id, dropdown = bc.dropdown(id_dict, options, label=label, **kwargs)
            self.outputs[drop_id] = "value"
            blocks.append([[dropdown]])

            # default = (
            #     "date" if "date" in df.columns else self.data[dataid]["columns"][0]
            # )
            # column = default
            # id_dict = {"name": f"{column}_datestart", **gid}
            # series = np.sort(df[column].unique().astype(int))
            # slider_id, slider = bc.slider(id_dict, series, label)
            # self.outputs[slider_id] = "value"
            # blocks.append([[slider]])

        block_size = {"ideal": [2, 3], "min": [1, 2]}
        return self._align(blocks, vertical, block_size)

    def date_slider(
        self, gid, dataid, column=None, variant="single", vertical=False, **kwargs
    ):
        df = self.data[dataid]["df"]
        default = "date" if "date" in df.columns else self.data[dataid]["columns"][0]
        column = column or default
        id_dict = {"name": f"{column}_filter", **gid}
        series = df[column].unique()
        if variant == "single":
            label = f"Select {column}:"
            slider_id, slider = bc.slider(
                id_dict, series, label, marks=True, variant="date"
            )
        elif variant == "range":
            label = f"Select range of {column}:"
            slider_id, slider = bc.range_slider(id_dict, series, label, marks=True)
        self.outputs[slider_id] = "value"

        blocks = [[[slider]]]
        block_size = {"ideal": [2, 4], "min": [1, 2]}
        return self._align(blocks, vertical, block_size)

    def date_combo(
        self, gid, dataid, column=None, variant="single", vertical=False, **kwargs
    ):
        df = self.data[dataid]["df"]
        default = "date" if "date" in df.columns else self.data[dataid]["columns"][0]
        column = column or default
        series = np.unique(df[column].dt.to_pydatetime())
        sl_series = np.unique(df[column])
        if variant == "single":
            id_dict = {"name": f"{column}_picker", **gid}
            label = f"Select {column}:"
            picker_id, picker = bc.date_picker(id_dict, series, label, variant=variant)
            id_dict = {"name": f"{column}_filter", **gid}
            slider_id, slider = bc.slider(id_dict, sl_series, None, marks=False)
        elif variant == "range":
            label = f"Select range of {column}:"
            picker_id, picker = bc.date_picker(id_dict, series, label)
        # self.outputs[picker_id] = "date"
        self.outputs[slider_id] = "value"

        blocks = [[[picker], [slider]]]
        block_size = {"ideal": [2, 2], "min": [1, 1]}
        return self._align(blocks, vertical, block_size)

    def date_picker(
        self, gid, dataid, column=None, variant="single", vertical=False, **kwargs
    ):
        df = self.data[dataid]["df"]
        default = "date" if "date" in df.columns else self.data[dataid]["columns"][0]
        column = column or default
        id_dict = {"name": f"{column}_filter", **gid}
        series = np.unique(df[column].dt.to_pydatetime())
        if variant == "single":
            label = f"Select {column}:"
            picker_id, picker = bc.date_picker(id_dict, series, label, variant=variant)
        elif variant == "range":
            label = f"Select range of {column}:"
            picker_id, picker = bc.date_picker(id_dict, series, label)
        self.outputs[picker_id] = "date"

        blocks = [[[picker]]]
        block_size = {"ideal": [2, 2], "min": [1, 1]}
        return self._align(blocks, vertical, block_size)

    def indicators(self, gid, dataid, components, vertical=False, **kwargs):
        blocks = []
        for ind_comp in components:
            name = ind_comp.get("name", ind_comp["args"]["y_column"])
            label = ind_comp.get("label", butil.titlize(name))
            id_dict = {"name": f"{name}_indicator", **gid}
            kwargs = {"Indicator.style": {"border": "1px solid gray"}, **kwargs}
            cid, comp = bc.indicator(id_dict, label=label, **kwargs)

            callback_name = f"{gid['pageid']}_{gid['bankid']}__update_{name}"
            callback_code = f"""
                idf = data["{dataid}"]["df"]
                inputs = dictutil.process_inputs(dash.callback_context.inputs)
                filters = butil.prepare_filters(inputs)
                sig, scale = butil.aggregate(idf, filters=filters, **{ind_comp['args']})
                sig = f'{{float(f"{{sig:.3g}}"):g}}'
                return f"{{sig}} {{scale}}{ind_comp.get("unit", "")}"
                """

            self.callbacks[cid] = {
                "provides": ["children"],
                "name": callback_name,
                "code": codeutil.format_code(callback_code),
            }

            blocks.append([[comp]])

        block_size = {"ideal": [2, 1.5], "min": [1, 1]}
        return self._align(blocks, vertical, block_size)

    def ranking(self, gid, dataid, **kwargs):
        id_dict = {"name": f"ranking", **gid}
        label = f"Top items"
        kwargs = {"Div.style": {"textAlign": "left"}, **kwargs}
        div_id, div = bc.div(id_dict, label=label, **kwargs)

        callback_name = f"{gid['pageid']}_{gid['bankid']}__update_ranking"
        callback_code = f"""
            idf = data["{dataid}"]["df"]
            inputs = dictutil.process_inputs(dash.callback_context.inputs)

            filters = butil.prepare_filters(inputs)
            fdf = butil.filter_df(idf, filters=filters)
            children = []

            column = dictutil.extract_unique("_column", inputs)
            key = dictutil.extract_unique("geo", inputs)
            inputs.update({{"key": key, "column": column}})
            inputs.update({kwargs})

            for item in butil.rank(fdf, **inputs):
                text = f"{{item[1]}}     {{item[0]}}"
                children.extend([html.Span(text), html.Hr()])
            return children
            """

        self.callbacks[div_id] = {
            "provides": ["children"],
            "name": callback_name,
            "code": codeutil.format_code(callback_code),
        }

        blocks = [[[div]]]
        block_size = {"ideal": [4, 2], "min": [1, 1]}
        return self._align(blocks, True, block_size)

    def filter_set(self, gid, dataid, columns=(), vertical=False, **kwargs):
        columns = columns or self.data[dataid]["keys"]
        blocks = []
        for col in columns:
            id_dict = {"name": f"{col}_filter", **gid}
            label = f"Select {col}".title()
            options = {
                "options": list(self.data[dataid]["df"][col].unique()),
                "default": [],
            }
            cargs = {"Dropdown.multi": True, **kwargs}
            drop_id, dropdown = bc.dropdown(id_dict, options, label=label, **cargs)
            self.outputs[drop_id] = "value"
            blocks.append([[dropdown]])

        # TODO Implement logical combinations once analytics treatment finished
        # if len(columns) >= 2:
        #     id_dict = {"name": f"filter_logic", **gid}
        #     cargs = {"Dropdown.clearable": False, **kwargs}
        #     options = {"options": ["And", "Or"], "default": "Or"}
        #     drop_id, dropdown = bc.dropdown(id_dict, options, label=None, **cargs)
        #     self.outputs[drop_id] = "value"
        #     blocks.append([[dropdown]])

        block_size = {"ideal": [2, 3], "min": [1, 2]}
        return self._align(blocks, vertical, block_size)

    def option_set(self, gid, components, vertical=True, **kwargs):
        blocks = []
        for comp in components:
            id_dict = {"name": comp["name"], **gid}
            label = comp["label"]
            # Default to dropdown
            ctype = bc.dropdown
            kwargs = {"Dropdown.clearable": False, **kwargs}
            options = comp["options"]
            if isinstance(comp["options"], str) and len(comp["options"] <= 3):
                ctype = bc.radio
            elif isinstance(comp["options"][0], (int, float)):
                ctype = bc.slider
                options = pd.Series(comp["options"])
            cid, optcomp = ctype(id_dict, options, label=label, marks=True, **kwargs)
            blocks.append([[optcomp]])
            self.outputs[cid] = "value"

        block_size = {"ideal": [3, 2], "min": [1, 1]}
        return self._align(blocks, vertical, block_size)

    def fit_controls(self, gid, dataid, vertical=False, **kwargs):
        blocks = []
        # Create the fit function dropdown
        id_dict = {"name": "fitfunc", **gid}
        options = ["linear", "quadratic"]
        label = "Fit Function"
        kwargs = {"Dropdown.clearable": False}
        fit_id, fit_comp = bc.dropdown(id_dict, options, label=label, **kwargs)
        self.outputs[fit_id] = "value"
        # Create the slider that sets training & prediction range
        id_dict = {"name": "fitrange", **gid}
        label = "Fit/Predict range: points to use, points to forecast"
        cid, fit_range = bc.range_slider(id_dict, options, label=label)
        self.outputs[cid] = "value"
        blocks.append([[fit_comp], [fit_range]])

        block_size = {"ideal": [2, 4], "min": [1, 2]}
        return self._align(blocks, vertical, block_size)

    def window_controls(self, gid, dataid, vertical=False, **kwargs):
        blocks = []
        id_dict = {"name": "window_transform", **gid}
        label = "Averaging Window"
        windows = [(1, "Day"), (7, "Week"), (30, "Month"), (365, "Year")]
        options = {
            "options": [{"value": item[0], "label": item[1]} for item in windows],
            "value": 1,
        }
        kwargs = {"Dropdown.clearable": False}
        cid, window = bc.dropdown(id_dict, options, label, **kwargs)
        self.outputs[cid] = "value"
        blocks.append([[window]])

        block_size = {"ideal": [2, 2], "min": [1, 1]}
        return self._align(blocks, vertical, block_size)

    def style_controls(self, gid, dataid, variants=(), vertical=False, **kwargs):
        blocks = []

        # A dropdown to select the line mode
        id_dict = {"name": "mode", **gid}
        options = ["lines+markers", "lines", "markers"]
        kwargs = {"Dropdown.clearable": False}
        mode_id, line_mode = bc.dropdown(id_dict, options, "Mode", **kwargs)
        self.outputs[mode_id] = "value"

        top_row = []
        if variants:
            id_dict = {"name": "variant", **gid}
            kwargs = {"Dropdown.clearable": False, **kwargs}
            label = "Select Graph Variant"
            var_id, variant = bc.dropdown(id_dict, variants, label, **kwargs)
            self.outputs[var_id] = "value"
            top_row.append(variant)

            # The radio buttons will be fed by the selected axis
            self.connectors[mode_id] = {
                "inputs": [(var_id, "value")],
                "outputs": [(mode_id, "options"), (mode_id, "value")],
            }

            callback_name = f"{gid['pageid']}_{gid['bankid']}__update_mode"
            callback_code = f"""
                inputs = dictutil.process_inputs(dash.callback_context.inputs)
                variant = dictutil.extract_unique("variant", inputs)
                options = ["lines+markers", "lines", "markers"]
                if variant == "bar":
                    options = ["group", "stack", "relative"]
                option_dict = butil.gen_options(options)
                return option_dict["options"], option_dict["value"]
                """

            self.callbacks[mode_id] = {
                "provides": ["options", "value"],
                "name": callback_name,
                "code": codeutil.format_code(callback_code),
            }
        top_row.append(line_mode)

        # Two sliders for the size of the markers and line width
        sizes = pd.Series([5, 20])
        id_dict = {"name": "marker_size", **gid}
        cid, marker_size = bc.slider(id_dict, sizes, "Marker Size", **kwargs)
        self.outputs[cid] = "value"
        sizes = pd.Series([1, 10])
        id_dict = {"name": "line_width", **gid}
        cid, line_width = bc.slider(id_dict, sizes, "Line Width", **kwargs)
        self.outputs[cid] = "value"
        bottom_row = [marker_size, line_width]

        block_size = {"ideal": [2, 4], "min": [1, 2]}
        blocks = [[top_row, bottom_row]]
        return self._align(blocks, vertical, block_size)

    def data_table(self, gid, dataid, vertical=True, **kwargs):
        id_dict = {"name": "columns", **gid}
        label = f"{dataid.upper()}: Choose Columns To Show"
        option_args = dictutil.extract(r"^options$|^default$", kwargs)
        options = {
            "options": self.data[dataid]["columns"] + self.data[dataid]["keys"],
            "default": self.data[dataid]["columns"][:3],
            **option_args,
        }
        kwargs = {"Dropdown.multi": True}
        drop_id, dropdown = bc.dropdown(id_dict, options, label=label, **kwargs)
        self.outputs[drop_id] = "value"

        id_dict = {"name": "table", **gid}
        table_id, datatable = bc.table(id_dict, label=None, **kwargs)

        # Columns displayed in the data table are selected here
        self.connectors[table_id] = {
            "inputs": [(drop_id, "value")],
            "outputs": [(table_id, "columns"), (table_id, "data")],
        }

        callback_name = f"{gid['pageid']}_{gid['bankid']}__update_table"
        callback_code = f"""
            fdf = data["{dataid}"]["df"]
            # Default inputs will be anything passed in 'args' of component
            inputs = {kwargs}

            # Override the default inputs with anything from the callback
            inputs.update(dictutil.process_inputs(dash.callback_context.inputs))
            dropdown_cols = dictutil.extract_unique("columns", inputs)

            # Use only first 1000 rows for performance
            fdf = fdf[:1000]
            columns = [{{"name": item, "id": item}} for item in dropdown_cols]
            return [columns, fdf.to_dict('records')]
            """

        self.callbacks[table_id] = {
            "provides": ["columns", "data"],
            "name": callback_name,
            "code": codeutil.format_code(callback_code),
        }

        block_size = {"ideal": [8, 12], "min": [4, 6]}
        blocks = [[[dropdown], [datatable]]]
        return self._align(blocks, vertical, block_size)

    # @logutil.loginfo(level='debug')
    def graph(self, gid, dataid, category="normal", vertical=False, **kwargs):
        id_dict = {"name": "graph", **gid}
        cid, component = bc.graph(id_dict, **kwargs)

        # Set dependent variable based on the type of graph
        if category == "map":
            dep_var = "z"
        else:
            dep_var = "y"

        callback_name = f"{gid['pageid']}_{gid['bankid']}__update_figure"
        callback_code = f"""
            fdf = data["{dataid}"]["df"]
            # NOTE Thinking this through
            # Default inputs will be anything passed in 'args' of component
            # inputs = {kwargs}
            inputs = {{}}

            # Override the default inputs with anything from the callback
            inputs.update(dictutil.process_inputs(dash.callback_context.inputs))

            # NOTE At least in some cases we want a manual override
            inputs.update({kwargs})

            filters = butil.prepare_filters(inputs)
            transforms = butil.prepare_transforms(inputs, dep_var="{dep_var}")
            figure = Graph.{category}(fdf,
                filters=filters,
                transforms=transforms,
                **inputs)
            figure.update_layout(classes.graph)
            style={{'visibility': 'visible'}}
            return figure, style
            """

        self.callbacks[cid] = {
            "provides": ["figure", "style"],
            "name": callback_name,
            "code": codeutil.format_code(callback_code),
        }
        block_size = {"ideal": [8, 12], "min": [4, 4]}
        blocks = [[[component]]]
        return self._align(blocks, vertical, block_size)



mars_page = {
    "title": "Mars Population",
    "subtitle": "Demonstrating graph options and style",
    "dataid": "mars",
    "banks": {
        "selector": {"type": "axis_controls", "args": {"use": "xy"}},
        "filters": {"type": "filter_set", "args": {}},
        "agg": {
            "type": "indicators",
            "width": 2,
            "args": {
                "components": [
                    {
                        "args": {
                            "y_column": "population",
                            "fixed_filters": {"or": {"date": ["2149-04-21"]}},
                        }
                    }
                ]
            },
        },
        "traces": {"type": "graph"},
        "style": {
            "type": "style_controls",
            "args": {"vertical": True, "variants": ["scatter", "bar"]},
        },
    },
    "layout": [["selector", "filters", "agg"], ["traces"], ["style"]],
    "connections": {
        "style": {"traces"},
        "selector": {"traces"},
        "filters": {"traces", "agg"},
    },
}

stock_page = {
    "title": "Stock Prices",
    "subtitle": "Demonstrating analytical features",
    "dataid": "stock",
    "banks": {
        "selector": {"type": "axis_controls", "args": {"use": "y", "default": "open"},},
        "filters": {"type": "filter_set", "args": {"vertical": True}},
        "window": {"type": "window_controls"},
        "traces": {
            "type": "graph",
            "args": {"x_column": "date", "x_scale": "date", "mode": "lines"},
        },
    },
    "layout": [["selector", "filters", "window"], ["traces"]],
    "connections": {
        "selector": {"traces"},
        "window": {"traces"},
        "filters": {"traces"},
    },
}

covid_page = {
    "title": "COVID-19",
    "subtitle": "Tracking case load",
    "dataid": "covid",
    "banks": {
        "agg": {
            "type": "indicators",
            "width": 2,
            "args": {"components": [{"args": {"y_column": "cases"}}]},
        },
        "date": {"type": "date_picker"},
        "top10": {"type": "ranking", "args": {"key": "state", "column": "cases"}},
        "mapsettings": {
            "type": "option_set",
            "args": {
                "components": [
                    {
                        "name": "geo",
                        "label": "Select geography",
                        "options": ["us_states", "us_counties"],
                    },
                ]
            },
        },
        "countymap": {
            "type": "graph",
            "width": 7,
            "args": {"category": "map", "variant": "choropleth", "z_column": "cases",},
        },
    },
    "layout": [["agg", "date", "mapsettings"], ["top10", "countymap"], ["trend"],],
    "connections": {
        "mapsettings": {"countymap", "trend"},
        "date": {"countymap", "top10", "agg"},
    },
}

oilngas_page = {
    "dataid": "oil",
    "banks": {
        "selector": {
            "width": 3,
            "type": "axis_controls",
            "args": {"use": "xy", "scale": False},
        },
        "daterange": {"width": 4, "type": "date_slider", "args": {"variant": "range"},},
        "filters": {
            "type": "filter_set",
            "args": {"columns": ["type", "status"], "vertical": True},
            "width": 4,
        },
        "aggregates": {
            "type": "indicators",
            "args": {
                "components": [
                    {"args": {"y_column": "wells"},},
                    {"args": {"y_column": "gas_produced"}, "unit": "cf"},
                    {"args": {"y_column": "oil_produced"}, "unit": "bbl"},
                    {"args": {"y_column": "water_produced"}, "unit": "bbl"},
                ],
            },
        },
        "barplot": {"type": "graph", "args": {"variant": "bar"}},
        "pieplot": {"type": "graph", "args": {"variant": "pie", "x_column": "type"}},
        "satellite": {
            "type": "graph",
            "width": 8,
            "args": {"category": "map", "variant": "scatter", "mapbox_center": "auto"},
        },
        "style": {"type": "style_controls"},
    },
    "layout": [
        ["daterange", "aggregates", "selector"],
        ["filters", "barplot"],
        ["pieplot", "satellite"],
    ],
    "connections": {
        "daterange": {"barplot", "aggregates", "pieplot",},
        "filters": {"barplot", "aggregates", "satellite",},
        "selector": {"barplot", "pieplot"},
    },
}

svm_page = {
    "dataid": "mars",
    "banks": {
        "selector": {"type": "axis_controls", "args": {"use": "xy", "vertical": True},},
        "data_settings": {
            "type": "option_set",
            "args": {
                "components": [
                    {
                        "name": "n_samples",
                        "options": [100, 200, 300, 400, 500],
                        "label": "Sample Size:",
                    },
                    {
                        "name": "noise",
                        "options": [0, 0.2, 0.4, 0.6, 0.8, 1.0],
                        "label": "Noise Level:",
                    },
                ]
            },
        },
        "heatmap": {"type": "graph"},
    },
    "layout": [["heatmap"]],
    "sidebar": ["selector", "data_settings"],
    "connections": {"data_settings": {"heatmap"}, "selector": {"heatmap"},},
}

data_page = {
    "banks": {
        "mars_table": {"type": "data_table", "args": {"dataid": "mars"}},
        "oil_table": {"type": "data_table", "width": 6, "args": {"dataid": "oil"}},
        "stock_table": {"type": "data_table", "args": {"dataid": "stock"}},
        "covid_table": {"type": "data_table", "width": 6, "args": {"dataid": "covid"}},
    },
    "layout": [["mars_table", "oil_table"], ["stock_table", "covid_table"]],
    "connections": {},
}


descriptor = {
    "name": "demonstration",
    "theme": "dark",
    "appbar": {"title": "Bento Demo", "subtitle": "A gallery of Dash recreations"},
    "data": {
        "covid": {"module": "bento.sample_data.covid"},
        "mars": {"module": "bento.sample_data.mars"},
        "oil": {"module": "bento.sample_data.oil"},
        "stock": {"module": "bento.sample_data.stock"},
    },
    "pages": {
        "covid": covid_page,
        "oilngas": oilngas_page,
        "mars": mars_page,
        "stock": stock_page,
        "data": data_page,
    },
}


def serve():
    """Both writes and serves the demo dashboard 
    It is used in the "bento-dash" console script supplied by setup.py
    """
    print("Running simple Bento demonstration...")
    import sys
    from bento import bento

    bento.Bento(descriptor).write()
    sys.path.append(".")

    from bento_app import app  # noqa

    app.run_server(host="0.0.0.0", port=8050, debug=False)


if __name__ == "__main__":
    from bento import bento

    bento.Bento(descriptor).write()



import math

from bento.common import logger, logutil  # noqa

logging = logger.fancy_logger(__name__, level=10)


# @logutil.loginfo(level='debug')
def apply_grid(page):
    gridsize = {"width": 12, "height": 8, "rowstart": 1}
    if page.get("sidebar"):
        gridsize["width"] = 10
        gridsize["rowstart"] = 3
    layout = page.get("layout")
    arr, bout = arrange(page["banks"], layout, gridsize)
    return {
        "layout": arr,
        "banks": bout,
        "sidebar": page.get("sidebar"),
        "intro": page.get("intro"),
        "title": page.get("title", ""),
        "subtitle": page.get("subtitle", ""),
    }


# TODO Figure out where and when we need this
def _rescale_ref(ref_sizes, grid_width=12, def_width=12):
    new_sizes = {}
    for name, val in ref_sizes.items():
        new_width = math.ceil(val[1] * grid_width / def_width)
        new_sizes[name] = [val[0], new_width]
    return new_sizes


# @logutil.loginfo(level='debug')
def arrange(banks, arrangement, grid):
    """Receives an ordered 2-d matrix of items and calculates default positions"""
    # Simply stack banks if there's no other info
    # TODO Implement a "Tetris" layout that packs in components based on size
    if not arrangement:
        arrangement = [[{"bankid": bankid}] for bankid in banks]

    # logging.warning(arrangement)
    banks_out = []
    # Give banks a position if none is supplied
    y_step = int(grid["height"] / len(arrangement))
    for y_idx, arr_row in enumerate(arrangement):
        if not arr_row:
            continue
        curr_y = y_idx * y_step
        x_step = int(grid["width"] / len(arr_row))
        for x_idx, bank in enumerate(arr_row):
            bankid = bank["bankid"]
            curr_x = x_idx * x_step + grid["rowstart"]
            # TODO Decide here
            # ref = _rescale_ref(banks[bankid]["sizing"], grid_width=grid["width"])
            ref = banks[bankid]["sizing"]
            bank.update(
                {
                    "width": ref["ideal"][1],
                    "slack": ref["ideal"][1] - ref["min"][1],
                    "ref": ref,
                    "position": banks[bankid].get("position", [curr_y, curr_x]),
                    "column": curr_x,
                    "row": curr_y,
                }
            )
            bank["slack_frac"] = bank["slack"] / bank["ref"]["min"][1]
            banks_out.append(bank)
        appease(banks, arr_row, grid)
        for bank, r_neighbor in zip(arr_row[:-1], arr_row[1:]):
            r_neighbor["column"] = bank["column"] + bank["width"]
    return arrangement, banks_out


# @logutil.loginfo(level='debug')
def appease(banks, row, grid):
    """Trims rows until they fit inside the grid, if possible"""
    # Add up the total width allocated for the row
    total_width = row[-1]["width"]
    total_slack = row[-1]["slack"]
    for bank, neighbor in zip(row[:-1], row[1:]):
        total_width += bank["width"]
        total_slack += bank["slack"]

    # Iterate through
    diff = grid["width"] - total_width
    while diff < 0 and total_slack > 0:
        ordered = sorted(row, key=lambda x: x["slack_frac"], reverse=True)
        # logging.info("%Slack ordering")
        # logging.info(ordered)
        bank = ordered[0]
        if bank["slack"] > 0:
            bank["width"] -= 1
            bank["slack"] -= 1
            bank["slack_frac"] = bank["slack"] / bank["ref"]["min"][1]
            total_slack -= 1
            diff += 1



import pandas as pd
import numpy as np
import math
import plotly.express as px

from collections import defaultdict
from bento.common import logger, logutil, dictutil  # noqa

logging = logger.fancy_logger(__name__)


def desnake(text):
    """Turns underscores into spaces"""
    return text.strip().replace("_", " ")


def titlize(text):
    return desnake(text.title())


def get_unit(num):
    scale = ["f", "n", "u", "m", "", "K", "M", "G", "T", "P"]
    offset = 4
    try:
        exp = min(9, offset + math.log10(num) / 3)
        if exp < 0:
            return num, ""
        exp = int(exp)
        sig = round(num / 10 ** ((exp - 4) * 3), 2)
        return sig, scale[exp]
    except Exception:
        return num, ""


# @logutil.loginfo(level="debug")
def prepare_transforms(inputs, dep_var="y"):
    transforms = []
    dep_column = dictutil.extract_unique(f"{dep_var}_column", inputs, pop=False)
    for key, value in dictutil.extract("_transform", inputs, pop=False).items():
        if not value:
            continue
        if "window" in key:
            transforms.append((dep_column, ["rolling", "mean"], [(value,), ()]))
        elif "calc" in key:
            if value == "Diff":
                transforms.append((dep_column, ["diff"], [()]))
            elif value == "Acceleration":
                transforms.append((dep_column, ["diff", "diff"], [(), ()]))
            elif value == "Cumulative":
                transforms.append((dep_column, ["cumsum"], [()]))
        elif "norm" in key:
            if value == "Max":
                transforms.append((dep_column, ["div"], [("trace.max",)]))
            elif value == "Other Series":
                transforms.append((dep_column, ["div"], [("ref.0",)]))

    return transforms


# @logutil.loginfo(level="debug")
def prepare_filters(inputs):
    filters = defaultdict(lambda: defaultdict(list))
    logic = dictutil.extract_unique("filter_logic", inputs, default="or").lower()
    for key, values in dictutil.extract("_filter", inputs, pop=False).items():
        col = key.replace("_filter", "")
        if not values:
            continue
        if not isinstance(values, list):
            values = [values]
        if "date" in key and len(values) == 2:
            filters["between"][col].extend(values)
        else:
            filters[logic][col].extend(values)
    return filters


# NOTE Currently used for pie charts and ranking
# @logutil.loginfo(level='debug')
def filter_df(idf, filters):
    odf = idf
    for logic, columns in filters.items():
        for column, values in columns.items():
            if "datetime" in str(type(values[0])):
                values = [np.datetime64(item) for item in values]
            if logic == "between":
                odf = odf[(odf[column] >= values[0]) & (odf[column] <= values[1])]
            elif logic == "or":
                odf = odf[odf[column].isin(values)]
            elif logic == "and":
                odf = odf[odf[column].isin(values)]

    return odf


def rank(idf, key, column, count=10, **kwargs):
    fdf = idf.groupby([key]).sum().reset_index()
    fdf = fdf.nlargest(count, column)
    return zip(fdf[key], fdf[column])


# NOTE Used for preparing the traces for graphs
# TODO Should combine this with filter_df/
# @logutil.loginfo(level='debug')
def prepare_traces(idf, filters, key_columns):
    # NOTE Brought over from figure callback, default multi-column approach
    # TODO Figure out how to determine default columns from df
    # column = self.data.get("keys", self.data["columns"][0])[0]
    # def_x_column = self.data["columns"][1]
    # idf.groupby(column).max().reset_index().nlargest(8, def_x_column)[column]
    idf["label"] = ""
    idf.name = ""
    traces = [idf]
    for logic, columns in filters.items():
        new_traces = []
        if logic == "between":
            for column, values in columns.items():
                for df in traces:
                    new = df[(df[column] >= values[0]) & (df[column] <= values[1])]
                    new.name = df.name
                    new_traces.append(new)
        elif logic == "or":
            for column, values in columns.items():
                for df in traces:
                    for value in values:
                        new = df[df[column] == value]
                        new.name = df.name
                        try:
                            new.name += " " + value
                        except Exception:
                            # try:
                            new.name += " " + pd.to_datetime(value).strftime("%Y-%m-%d")
                            # except Exception:
                            #     logging.warning(f"Can't add {column} to trace name")
                        new_traces.append(new)
        elif logic == "and":
            for column, values in columns.items():
                for df in traces:
                    for value in values:
                        new = df[df[column] == value]
                        new.name = df.name
                        try:
                            new.name += " " + value
                        except Exception:
                            # try:
                            new.name += " " + pd.to_datetime(value).strftime("%Y-%m-%d")
                            # except Exception:
                            #     logging.warning(f"Can't add {column} to trace name")
                        new_traces.append(new)

        traces = new_traces

    new_traces = []
    for df in traces:
        new = df.groupby(key_columns).sum().reset_index()
        new.name = df.name
        new_traces.append(new)
    traces = new_traces

    return traces


# @logutil.loginfo(level="debug")
def trace_analytics(traces, transforms):
    for transform in transforms:
        column, operations, arg_list = transform
        for trace in traces:
            buff = trace[column]
            for op, args in zip(operations, arg_list):
                final_args = []
                for arg in args:
                    if "trace" in str(arg):
                        oper = arg.split(".")[-1]
                        final_args.append(getattr(buff, oper)())
                    elif "ref" in str(arg):
                        idx = 0
                        final_args.append(traces[idx][column])
                    else:
                        final_args.append(arg)
                buff = getattr(buff, op)(*final_args)
            trace[column] = buff
    return traces


def aggregate(idf, y_column=None, filters=None, logic="sum", keys=None, **kwargs):
    filters = filters or {}
    filters.update(kwargs.get("fixed_filters", {}))
    # TODO Plenty of work to do cleaning up the data processing utilities like this
    keys = keys or ["date"]
    traces = prepare_traces(idf, filters, keys)
    agg_df = pd.concat(traces)
    if not y_column:
        return len(agg_df), ""
    quantity = getattr(agg_df[y_column], logic)()
    return get_unit(quantity)


def _date_marks(ordered):
    spacing = 7
    style = {}
    labels = {item: pd.Timestamp(item).date().day for item in ordered[::spacing]}
    marks = {
        int(key): {"label": label, "style": style} for key, label in labels.items()
    }
    return marks


def gen_marks(series, variant="auto"):
    """Processes a dataframe column into a valid slider series"""
    ordered = sorted(series)
    spacing = math.ceil(len(ordered) / 10)
    if variant == "date":
        marks = _date_marks(ordered)
    else:
        marks = {item: str(item) for item in series[::spacing]}
    return marks


def gen_options(option_input):
    # In this case, we're given just the set of options only, assuming first is default
    if isinstance(option_input, list):
        option_list = option_input
        default = option_list[0]
    # The default may be specified in the dict version
    elif isinstance(option_input, dict):
        # TODO Make this more robust
        if "value" in option_input:
            return option_input
        option_list = option_input["options"]
        default = option_input.get("default", option_list[0])
    # TODO Can we determine when we should run desnake on the entries?
    options = [{"label": desnake(item).title(), "value": item} for item in option_list]
    return {"options": options, "value": default}


def get_first_numeric(data_types):
    for dtype in data_types:
        if "log" in data_types[dtype]:
            return dtype


def log_color_scale(name, base=2.718, category="sequential"):
    color_category = getattr(px.colors, category)
    color_sequence = getattr(color_category, name)
    # print(color_sequence)
    log_val = [round(1 / base ** idx, 10) for idx in range(len(color_sequence))][::-1]
    log_val[0] = 0
    log_sequence = list(zip(log_val, color_sequence))
    return log_sequence


# @logutil.loginfo(level='debug')
def data_range(pd_df_list, column, scale="category"):
    minimum = pd_df_list[0][column].min()
    maximum = pd_df_list[0][column].max()
    for df in pd_df_list:
        series = df[column]
        minimum = min(minimum, series.min())
        maximum = max(maximum, series.max())
        if scale == "log":
            minimum = max(0.1, minimum)
            minimum = np.floor(max(-1, np.log10(minimum)))
            maximum = np.ceil(np.log10(maximum))
        elif scale == "linear":
            minimum = np.floor(minimum)
            maximum = np.ceil(maximum)
    return [minimum, maximum]


if __name__ == "__main__":

    log_color_scale("Viridis", base=10)

    for i in range(-18, 19):
        unit = "Hz"
        sig, scale = get_unit(10 ** i)
        print(f"{sig} {scale}{unit}")



# Page and bank ids, are strings of word characters only with length 2+
# No trailing, leading, or double underscores
bento_uid_regex = r"^(?!_|.*_$|.*__.*)[a-z0-9_]{2,}$"

# One or more words separated by a space
words_regex = r"^\w+( \w+)*$"

page_schema = {
    "banks": {
        "type": "dict",
        "required": True,
        "allow_unknown": True,
        "minlength": 1,
        "keysrules": {"type": "string", "regex": bento_uid_regex},
        "valuesrules": {
            "type": "dict",
            "allow_unknown": True,
            "schema": {
                "type": {"type": "string", "required": True, "regex": bento_uid_regex},
                "width": {"type": "integer"},
                "args": {"type": "dict"},
            },
        },
    },
    "connections": {"type": "dict"},
    "dataid": {"type": "string"},
    "layout": {"type": "list"},
    "title": {"type": "string"},
    "subtitle": {"type": "string"},
}

descriptor_schema = {
    "name": {"type": "string"},
    "theme": {"type": "string", "regex": words_regex},
    "theme_dict": {"type": "dict"},
    "appbar": {
        "type": "dict",
        "schema": {
            "title": {"type": "string"},
            "subtitle": {"type": "string"},
            "image": {"type": "string"},
        },
    },
    "data": {"type": "dict"},
    "pages": {
        "type": "dict",
        "allow_unknown": True,
        "required": True,
        "minlength": 1,
        "keysrules": {"type": "string", "regex": bento_uid_regex},
        "valuesrules": {"type": "dict", "schema": page_schema},
    },
}



import flask
import functools
import inspect
import os
import traceback
from bento.common import logger

# logging = logger.fancy_logger(__name__)
# We want to use a simple format here because we care about the wrapped function
# Not the location and name of the wrapper
logging = logger.fancy_logger(__name__, fmt="bare")


def _parse_id(func):
    """Supplies the module name and functon name as a 2-tuple"""
    return (func.__module__.split(".")[-1], func.__name__)


def loginfo(level="debug", log_route=False):
    def inner(func):
        """Wraps class methods to provide automatic log output on args/return value"""
        module, name = _parse_id(func)

        # Retrieves list of ordered arg names from the decorated function
        arg_names = list(inspect.signature(func).parameters.keys())
        class_method = False
        # Removes class-related leading arg from consideration
        if arg_names and arg_names[0] in ("self", "cls"):
            class_method = True
            arg_names = arg_names[1:]

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Look back in the call stack for the appropriate function
            stack_idx = -2
            stack = traceback.extract_stack()
            lineno = stack[stack_idx].lineno
            callFunc = stack[stack_idx].name
            filename = os.path.basename(stack[stack_idx].filename).split(".")[0]
            msg_base = f"{module}.{name} @ {filename}.{callFunc}:{lineno}"

            # Log each argument to the function
            out_args = {}
            start_idx = 0
            if class_method:
                start_idx = 1
            for arg_name, arg_value in zip(arg_names, args[start_idx:]):
                out_args[arg_name] = arg_value
            out_args.update(kwargs)
            msg = f"{msg_base} - Input Arguments:"
            getattr(logging, level)(msg)
            getattr(logging, level)(out_args)

            # The call itself
            result = func(*args, **kwargs)

            # If flagged, also try to log the request object
            if log_route:
                logging.debug(flask.request.json)

            # Similarly log the result of the function
            if type(result) in (dict, list, tuple):
                typename = type(result).__name__
                msg = f"{msg_base} - Returning {typename} of length {len(result)}"
                getattr(logging, level)(msg)
            getattr(logging, level)(result)
            return result

        return wrapper

    return inner



import json
import pathlib
import pkgutil
from urllib import request

from bento.common import logger, dictutil  # noqa

logging = logger.fancy_logger(__name__)

# TODO Generalize resource loading
init_py_path = pkgutil.get_loader("bento").path
package_path = pathlib.Path(init_py_path).parent

geojson = {}
resource_list = [
    ("us_counties", "geojson_us_counties.json"),
    ("us_states", "geojson_us_states.json"),
    ("world", "geojson_world.json"),
]

web_backup = {
    "us_counties": (
        "https://raw.githubusercontent.com/plotly/"
        "datasets/master/geojson-counties-fips.json"
    ),
    "us_states": (
        "https://raw.githubusercontent.com/PublicaMundi/"
        "MappingAPI/master/data/geojson/us-states.json"
    ),
}
location_list = [".", "assets", f"{package_path}/assets"]
for uid, resource in resource_list:
    for loc in location_list:
        try:
            with open(f"{loc}/{resource}", "r") as fh:
                geojson[uid] = json.load(fh)
            logging.info(f"...Loaded {resource}")
            break
        except Exception:
            continue

    if uid in web_backup and uid not in geojson:
        logging.info("   Trying web...")
        try:
            with request.urlopen(web_backup[uid]) as response:
                geojson[uid] = json.load(response)
                logging.info(f"...Loaded {resource} from web")
        except Exception:
            logging.warning(f"   ...Failed to load geojson[{uid}]")



gray = {}
gray[100] = "#f8f9fa"
gray[200] = "#ecf0f1"
gray[300] = "#dee2e6"
gray[400] = "#ced4da"
gray[500] = "#b4bcc2"
gray[600] = "#95a5a6"
gray[700] = "#7b8a8b"
gray[800] = "#343a40"
gray[900] = "#212529"

blue = "#133C93"
indigo = "#6610f2"
purple = "#6f42c1"
pink = "#e83e8c"
red = "#e74c3c"
orange = "#fd7e14"
yellow = "#f39c12"
green = "#18bc9c"
teal = "#20c997"
cyan = "#3498db"

# This the default 'Bento' theme
# NOTE Still plenty of work to do fleshing this out (variants, secondaries)
bento = {
    # Colors
    "color__primary": blue,
    "color__secondary": purple,
    "color__on_primary": gray[200],
    "color__on_surface": gray[800],
    "color__on_surface_secondary": gray[700],
    "color__background": gray[100],
    "color__info": cyan,
    "color__warning": yellow,
    "color__danger": red,
    "color__success": green,
    "color__surface": gray[300],
    "color__plot_bg": gray[500],
    # Density
    "pad_unit": 4,
    "pad_pct": 1.0,
    # Flatness
    "elevation": 4,
    # Font
    "font__family": ["Lato", "Roboto", "sans-serif"],
    "font__size": "1.5rem",
    # Map
    "map__mapbox_style": "carto-positron",
}



from bento.common import logger, util

logging = logger.fancy_logger(__name__)


def load():
    filename = "sample_oilngas_data.csv"
    data = {
        "df": util.df_loader(filename, parse_dates=False),
        "keys": ["county", "type", "status"],
        "types": {
            "date": int,
            "wells": int,
            "water_produced": float,
            "oil_produced": float,
            "gas_produced": float,
        },
    }
    return data



"""These are a set of functions that wrap and facilitate use of dash components.
A list of goals for these functions:
    * Allow easy templating of a user's intended Dash app
    * Provide convenient default values
    * Don't do anything irreversible
"""
import math

from bento.common import logger, logutil, dictutil  # noqa
from bento import util as butil

logging = logger.fancy_logger(__name__)


def _create(component_class, id_dict, args, lib="dcc", suffix="", label=None, **kwargs):
    """Helps wrap Dash components so they can be easily written out via Jinja2"""
    bankid, name, pageid = dictutil.pluck(id_dict)
    cid = f"{pageid}/{bankid}|{name}{suffix}"
    comp = {
        "lib": lib,
        "component": component_class,
        "label": label,
        "args": {
            "id": cid,
            **args,
            # Allows users to supply arguments to Dash components via the descriptor
            **dictutil.extract_path(f"{component_class}.", kwargs),
        },
    }
    return cid, comp


# --- The rest are all component wrappers ---
# @logutil.loginfo(level='debug')
def graph(id_dict, **kwargs):
    # TODO Flesh this out
    # This sets the graphs to take up space but not be visible prior to callback
    args = {"style": {"visibility": "hidden"}}
    if "height" in kwargs:
        args["style"]["height"] = kwargs["height"]
    return _create("Graph", id_dict, args, **kwargs)


def radio(id_dict, options, label=None, **kwargs):
    args = {**butil.gen_options(options)}
    return _create("RadioItems", id_dict, args, label=label, **kwargs)


def dropdown(id_dict, options, label=None, **kwargs):
    args = {**butil.gen_options(options)}
    return _create("Dropdown", id_dict, args, label=label, **kwargs)


def slider(id_dict, series, label=None, marks=False, variant="auto", **kwargs):
    # TODO Use dict extraction
    args = {
        "value": kwargs.get("value") or series.min(),
        "min": series.min(),
        "max": series.max(),
    }
    if variant == "date":
        args = {
            "value": kwargs.get("value") or int(series.min()),
            "min": int(series.min()),
            "max": int(series.max()),
        }

    if marks:
        args["marks"] = butil.gen_marks(series, variant)
        args["step"] = None
    return _create("Slider", id_dict, args, label=label, **kwargs)


def range_slider(id_dict, series, label=None, marks=None, **kwargs):
    spacing = math.ceil(len(series) / 10)
    args = {
        "min": series.min(),
        "max": series.max(),
        "value": [series.min(), series.max()],
        "marks": {str(item): str(item) for item in sorted(series)[::spacing]},
    }
    return _create("RangeSlider", id_dict, args, label=label, **kwargs)


def date_picker(id_dict, series, label=None, variant="single", **kwargs):
    args = {
        "min_date_allowed": series.min(),
        "max_date_allowed": series.max(),
        "initial_visible_month": series.max(),
        "date": series.max(),
    }
    return _create("DatePickerSingle", id_dict, args, label=label, **kwargs)


def stepper(id_dict, **kwargs):
    args = {
        "disabled": True,
        "interval": 0,
        "n_intervals": 0,
    }
    return _create("Interval", id_dict, args, label=None, **kwargs)


def table(id_dict, label=None, **kwargs):
    args = {
        "page_action": "none",
        "filter_action": "native",
        "sort_action": "native",
        "style_table": {"overflowX": "auto", "overflowY": "auto", "height": "300px",},
        # NOTE This is needed until Plotly fixes name collision with Bootstrap Grid css
        "css": [{"selector": ".row-1", "rule": "margin-right: 0px; margin-left: 0px"},],
    }
    return _create("DataTable", id_dict, args, lib="dash_table", label=label, **kwargs)


def indicator(id_dict, label=None, **kwargs):
    args = {"children": "<Indicator>"}
    return _create("H3", id_dict, args, lib="html", label=label, **kwargs)


def div(id_dict, label=None, **kwargs):
    args = {"children": "<Div>"}
    return _create("Div", id_dict, args, lib="html", label=label, **kwargs)



# import pandas as pd
# import numpy as np
# import datetime as dt
import math
import plotly.graph_objects as go

from bento import util as butil
from bento.resources import geojson
from bento.common import logger, dictutil  # noqa

logging = logger.fancy_logger(__name__)


class Graph:
    @classmethod
    def normal(
        cls,
        idf,
        variant="scatter",
        x_column=None,
        y_column=None,
        x_label=None,
        y_label=None,
        x_scale="linear",
        y_scale="linear",
        color=None,
        opacity=0.7,
        mode="lines+markers",
        line_width=3,
        marker_size=10,
        marker_line_width=0.5,
        marker_line_color="black",
        keys=None,
        filters=[],
        transforms=[],
        **kwargs,
    ):

        x_label = x_label or x_column

        if isinstance(y_column, str):
            y_columns = [y_column]
            y_label = y_label or y_column
        elif isinstance(y_column, list):
            y_columns = y_column
            y_label = y_label or y_column[0]

        key_columns = keys or ["date"]

        graph_call = getattr(go, variant.title())

        fig = go.Figure()
        if variant in ("pie"):
            fdf = butil.filter_df(idf, filters)
            default_settings = {}
            data_settings = {
                "pie": {"labels": fdf[x_column], "values": fdf[y_column],},
            }
            style_settings = {}
            settings = data_settings.get(variant, default_settings)
            settings.update(style_settings.get(variant, {}))

            fig.add_trace(graph_call(**settings))

        elif variant in ("scatter", "bar", "histogram"):
            traces = butil.prepare_traces(idf, filters, key_columns)
            traces = butil.trace_analytics(traces, transforms)
            for trace_df in traces:
                y_idx = 1
                for y_column in y_columns:
                    yaxis = f"y{y_idx}"

                    default_settings = {
                        "x": trace_df[x_column],
                        "y": trace_df[y_column],
                        # TODO Fix up hover info for non-map plots
                        # "text": trace_df["hover_info"],
                        "name": f"{trace_df.name} - {y_column.title()}",
                    }

                    if y_idx > 1:
                        default_settings.update({"yaxis": yaxis})

                    data_settings = {
                        "histogram": {"x": trace_df[y_column]},
                    }

                    style_settings = {
                        "scatter": {
                            "opacity": opacity,
                            "mode": mode,
                            "line_width": line_width,
                            "marker_size": marker_size,
                            "marker_line_width": marker_line_width,
                            "marker_line_color": marker_line_color,
                        },
                    }

                    settings = data_settings.get(variant, default_settings)
                    settings.update(style_settings.get(variant, {}))

                    # TODO An early way to introduce using marker color that needs work
                    if color:
                        settings["marker_color"] = (
                            trace_df[color].astype("category").cat.codes
                        )
                        settings["text"] = trace_df[color]

                    fig.add_trace(graph_call(**settings))
                    y_idx += 1

        barmode = mode if mode in ["stack", "group", "relative"] else "stack"
        layout = {
            "margin": {
                "l": 16 + 4 * 4,
                "b": 40 + 4 * 4,
                "t": 40 + 4 * 4,
                "r": 40 + 4 * 4,
            },
            "barmode": barmode,
            "legend_x": 0,
            "legend_y": 1,
            "legend_bgcolor": "rgba(0,0,0,0)",
            "legend_font_size": 10,
            "xaxis": {"type": x_scale, "title": butil.titlize(x_label)},
            "yaxis": {"type": y_scale, "title": butil.titlize(y_label)},
        }

        # TODO Temporary multi-axis support
        for y_idx in range(1, len(y_columns)):
            layout.update(
                {
                    f"yaxis{y_idx+1}": {
                        "type": y_scale,
                        "title": y_columns[y_idx],
                        "overlaying": "y",
                        "side": "right",
                    }
                }
            )

        if variant == "pie":
            layout.update({"legend_x": 1, "legend_y": 1})

        if variant == "histogram":
            layout = {
                "xaxis": {"title": butil.titlize(y_label),},
                "yaxis": {"title": f"Histogram Count of {y_label}",},
            }

        fig.update_layout(layout)
        return fig

    @classmethod
    def map(
        cls,
        idf,
        variant="Scatter",
        z_column=None,
        geo="us_states",
        mapbox_center="default",
        mapbox_style="carto-positron",
        mapbox_zoom=3,
        marker_size=5,
        marker_opacity=0.8,
        marker_line_width=0,
        marker_line_color="black",
        filters=[],
        **kwargs,
    ):
        fig = go.Figure()

        # First, create the single data trace
        base_args = {
            "marker_opacity": marker_opacity,
            "name": getattr(idf, "name", ""),
        }
        pdf = butil.filter_df(idf, filters)

        if variant == "scatter":
            pdf = pdf.groupby(["latitude", "longitude"]).sum().reset_index()
            hovertemplate = "<b>Loc</b><br>Latitude: %{lat}<br>Longitude: %{lon}<br>"
            args = {
                **base_args,
                "lon": pdf["longitude"],
                "lat": pdf["latitude"],
                "marker_size": marker_size,
                "hovertemplate": hovertemplate,
            }
        elif variant == "choropleth":
            # Leading two digits of fips are state code
            if "state" in geo:
                geo = "us_states"
                pdf.loc[:, "fips"] = pdf["fips"].astype(str).str.slice(0, 2)
                pdf = pdf.groupby(["fips", "state"]).sum().reset_index()
                loc_column = "fips"
                text = pdf["state"]
            elif "count" in geo:
                geo = "us_counties"
                loc_column = "fips"
                text = pdf["county"]
            elif "world" in geo:
                geo = "world"
                loc_column = "alpha3"
                text = pdf["country"]

            ht_title = "<b>%{text}</b>"
            ht_info = f"{z_column.title()}: %{{z:s}}"
            hovertemplate = "<br>".join([ht_title, ht_info])
            args = {
                **base_args,
                "z": pdf[z_column],
                "geojson": geojson[geo],
                "text": text,
                "locations": pdf[loc_column],
                "marker_line_width": marker_line_width,
                "marker_line_color": marker_line_color,
                "colorscale": butil.log_color_scale("Viridis", base=3),
                "showscale": False,
                "hovertemplate": hovertemplate,
            }

        trace = getattr(go, f"{variant.capitalize()}mapbox")(args)
        fig.add_trace(trace)

        if mapbox_center == "default":
            if "us" in geo:
                mapbox_center = {"lat": 37.0902, "lon": -95.7129}
                mapbox_zoom = 3
            elif "world" in geo:
                mapbox_center = {"lat": 0, "lon": 0}
                mapbox_zoom = 1
        elif mapbox_center == "auto":
            if variant == "scatter":
                ref_lat, ref_lon, ref_zoom = 25, 60, 3
                midpoint_lat = (pdf["latitude"].min() + pdf["latitude"].max()) / 2
                midpoint_lon = (pdf["longitude"].min() + pdf["longitude"].max()) / 2
                mapbox_center = {"lat": midpoint_lat, "lon": midpoint_lon}

                lat_delta = pdf["latitude"].max() - pdf["latitude"].min()
                lon_delta = pdf["longitude"].max() - pdf["longitude"].min()
                lat_multiple = ref_lat / lat_delta
                lon_multiple = ref_lon / lon_delta
                magnification = min(lat_multiple, lon_multiple)
                mapbox_zoom = ref_zoom + math.log(magnification, 2)

        # Now define the layout
        layout = {
            "margin": {"l": 0, "b": 0, "t": 0, "r": 0},
            "mapbox_style": mapbox_style,
            "mapbox_zoom": mapbox_zoom,
            "mapbox_center": mapbox_center,
        }
        fig.update_layout(layout)
        return fig



import pathlib
from setuptools import setup

# The directory containing this file, relative to cwd
HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()


def version():
    _version = {}
    with open("bento/_version.py") as fh:
        exec(fh.read(), _version)
        return _version["__version__"]


def requires(req_mode):
    with open("requires-{}.txt".format(req_mode)) as fh:
        requires = [line.strip() for line in fh]
        return [req for req in requires if req and not req.startswith("#")]


setup(
    name="bento-dash",
    version=version(),
    description="Create Plotly Dash apps via templates",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/dereklarson/bento",
    author="Derek Larson",
    author_email="larson.derek.a@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
    packages=["bento", "bento.common", "bento.dashboards", "bento.sample_data"],
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=requires("install"),
    entry_points={"console_scripts": ["bento-demo=bento.dashboards.demo:serve"]},
)



from bento.common import logger, util

logging = logger.fancy_logger(__name__)


def load():
    filename = "sample_mars_data.csv"
    data = {
        "df": util.df_loader(filename, parse_dates=False),
        "keys": ["region", "city"],
        "types": {"date": "date", "population": int, "energy_consumption": float,},
    }
    return data



import re

from bento.common import logutil  # noqa
from bento.common.logger import fancy_logger

logging = fancy_logger(__name__)


def pluck(input_dictionary):
    """Destructures a dictionary (vaguely like ES6 for JS)
       Results are SORTED by key alphabetically! You must sort your receiving vars
    """
    items = sorted(input_dictionary.items())
    return [item[1] for item in items]


def _cid2c(cid):
    """Gets just the component portion of a cid string
    e.g. main_page/axis_controls|x_column.value => x_column
    """
    return cid.split("|")[-1].split(".")[0]


def process_inputs(input_dictionary):
    """Alters the keys of a dictionary by a helper function"""
    return {_cid2c(key): val for key, val in input_dictionary.items()}


def extract(regex, input_dictionary, unique=False, pop=True):
    """Splits off a subset dictionary with keys matching the provided 'path' prefix"""
    method = "pop" if pop else "get"
    match_keys = [key for key in input_dictionary if re.search(regex, key)]
    output = {key: getattr(input_dictionary, method)(key) for key in match_keys}
    return output


def extract_unique(regex, input_dictionary, pop=True, default=None):
    method = "pop" if pop else "get"
    match_keys = [key for key in input_dictionary if re.search(regex, key)]
    output = {key: getattr(input_dictionary, method)(key) for key in match_keys}

    if len(match_keys) > 1:
        logging.warning(f"Regex {regex} not unique: {len(match_keys)} matches")
        logging.warning(input_dictionary)
    elif len(match_keys) == 0:
        return default
    else:
        return pluck(output)[0]


def extract_path(path, input_dictionary):
    """Splits off a subset dictionary with keys matching the provided 'path' prefix"""
    match_keys = [key for key in input_dictionary if key.startswith(path)]
    output = {key.replace(path, ""): input_dictionary.pop(key) for key in match_keys}
    return output


def merge(d_base, d_in, loc=None):
    """Adds leaves of nested dict d_in to d_base, keeping d_base where overlapping"""
    loc = loc or []
    for key in d_in:
        if key in d_base:
            if isinstance(d_base[key], dict) and isinstance(d_in[key], dict):
                merge(d_base[key], d_in[key], loc + [str(key)])
            elif isinstance(d_base[key], list) and isinstance(d_in[key], list):
                for item in d_in[key]:
                    d_base[key].append(item)
            elif d_base[key] != d_in[key]:
                d_base[key] = d_in[key]
        else:
            d_base[key] = d_in[key]
    return d_base


# @logutil.loginfo()
def flatten(dict_in, delim="__", loc=[]):
    """Un-nests the dict by combining keys, e.g. {'a': {'b': 1}} -> {'a_b': 1}"""
    loc = loc or []
    output = {}
    if not dict_in and loc:
        output[delim.join(loc)] = {}
    for key in dict_in:
        if isinstance(dict_in[key], dict):
            nest_out = flatten(dict_in[key], delim=delim, loc=loc + [str(key)])
            output.update(nest_out)
        else:
            base_key = delim.join(loc + [str(key)])
            output[base_key] = dict_in[key]
    return output


def nest(dict_in, delim="__"):
    """Nests the input dict by splitting keys (opposite of flatten above)"""
    # We will loop through all keys first, and keep track of any first-level keys
    # that will require a recursive nest call. 'renest' stores these keys
    output, renest = {}, []
    for key in dict_in:
        if delim not in key:
            output[key] = dict_in[key]
            continue
        loc = key.split(delim, 1)
        if loc[0] not in output:
            output[loc[0]] = {}
            # Uniquely add the key to the subdictionary we want to renest
            renest.append(loc[0])
        output[loc[0]][loc[1]] = dict_in[key]

    # Renest all higher-level dictionaries
    for renest_key in renest:
        output[renest_key] = nest(output[renest_key])
    return output


def common_keys(dict_a, dict_b):
    """Returns new dict with the common keys of the input dicts using dict_a values"""
    return {key: dict_a[key] for key in dict_a.keys() & dict_b.keys()}


class Edict(dict):
    def __init__(self, initial):
        if not hasattr(initial, "__iter__"):
            try:
                # This works fine for dataclasses, it seems
                super().__init__(vars(initial))
            except Exception:
                logging.debug("Failed to ingest into Edict...")
                logging.debug(initial)
        else:
            super().__init__(initial)

    def __and__(self, other):
        """Returns the first dict with only keys that overlap with the second"""
        return common_keys(self, other)



import pathlib
import pkgutil
import re
import subprocess
import pandas as pd

from bento.common import logger

logging = logger.fancy_logger(__name__)


def id_func(x):
    return x


def desnake(text):
    """Turns underscores into spaces"""
    return text.strip().replace("_", " ")


def snakify(text):
    for char in "()":
        text = text.replace(char, "")
    text = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", text).strip()
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", text).lower().replace(" ", "_")


def snakify_column_names(df):
    col_map = {col_name: snakify(col_name) for col_name in df.columns}
    return df.rename(columns=col_map)


def nice_command(cmd):
    proc = subprocess.Popen(cmd)
    try:
        logging.info(f"Waiting on {cmd[0]}")
        logging.debug(cmd)
        proc.wait()
    except KeyboardInterrupt:
        logging.info(f"Letting {cmd[0]} clean up...")
        proc.wait()
        logging.info("...Done")


def df_loader(filename, package="bento", parse_dates=["date"], location="."):
    args = {
        "index_col": 0,
        "parse_dates": parse_dates or [],
        "infer_datetime_format": True,
    }
    # First try locally for an override file, then check assets
    init_py_path = pkgutil.get_loader(package).path
    package_path = pathlib.Path(init_py_path).parent
    location_list = [location, "assets", f"{package_path}/assets"]
    for loc in location_list:
        try:
            df = pd.read_csv(f"{loc}/{filename}", **args)
            # TODO Figure out some log-based way to get this output cleanly
            # logging.info(f"*** Loaded DF from {filename} with {len(df)} rows***")
            # if logging.level <= 10:
            #     print(df.head(3))
            return df
        except FileNotFoundError:
            logging.debug(f"Didn't find {filename} at {loc}")

    logging.warning(f"Unable to load {filename} from any of {location_list}")


# Runs a supplied shell command, handling output
def logged_command(cmd, output_dir=".logs", shell=False):
    subprocess.run(["mkdir", "-p", output_dir])
    base_cmd = cmd[0]
    logging.info(f"Running command: {cmd}")
    output = ""
    errors = ""
    with open(f"{output_dir}/{base_cmd}_std.log", "w+") as fout:
        with open(f"{output_dir}/{base_cmd}_err.log", "w+") as ferr:
            result = subprocess.run(cmd, stdout=fout, stderr=ferr, shell=shell)
            try:
                fout.seek(0)
                output = fout.read()
                ferr.seek(0)
                errors = ferr.read()
                if errors:
                    logging.debug("Stderr: {}".format(errors))
            except Exception as exc:
                logging.warning("Issue with capturing log output of subprocess")
                logging.info(exc)

    # TODO refactor with dataclass or other?
    cmd_result = {
        "code": result.returncode,
        "out": output,
        "err": errors,
    }
    return cmd_result


# TODO Look at replacing this with functools.lru_cache
# Decorator: makes functions perform better by reusing cached results for same args
def memoize(func):
    cache = func.cache = {}

    def run(*args, **kwargs):
        # NOTE This does break if args contains non-hashable objects like lists
        key = args + tuple(sorted(kwargs.items()))
        if key in cache:
            return cache[key]
        else:
            result = func(*args, **kwargs)
            cache[key] = result
            return result

    return run



import black
import cerberus
import copy
import importlib
import pathlib
import re
from jinja2 import Environment, PackageLoader

from bento import util as butil
from bento import banks, grid, schema, style

from bento.common import logger, logutil, dictutil, codeutil  # noqa

logging = logger.fancy_logger(__name__, fmt="simple")


class Bento:
    def __init__(self, descriptor, init_only=False):
        """Converts the supplied descriptor into a jinja2 template context"""
        self.template = "bento_v1.py.j2"
        self.baseline_template = "baseline.css.j2"
        self.theme_template = "theme.css.j2"

        # Catches any problems with the input descriptor up front
        if not self.is_valid(descriptor):
            logging.warning("Exiting Bento intialization after failed validation")
            return

        # Stores a normalized version of the descriptor, ensuring uniformity
        self.desc = self.normalize(descriptor)

        # Loads the input data to inform components to the columns, types, etc
        self.data = self.process_data(self.desc)
        self.bb = banks.BentoBanks(self.data)

        # Generates the initial context object
        self.init_structure()

        if init_only:
            return

        logging.info("Creating the pages:")
        for pageid, page in self.desc["pages"].items():
            logging.info(f"#^  {pageid}...")
            self.create_page(pageid, page)
            # Specifies and attaches connectors to callbacks
            self.connect_page(page)
            logging.info(f"#$+ connected")

    # @logutil.loginfo(level='debug')
    def is_valid(self, descriptor):
        """Ensures a proper naming scheme is followed by the IDs"""
        logging.info("#^Testing validity of input descriptor...")
        validator = cerberus.Validator(schema.descriptor_schema)
        self.valid = validator.validate(descriptor)
        if not self.valid:
            logging.info("#$- failed")
            logging.warning("Descriptor validaiton failed due to the following errors:")
            logging.warning(validator.errors)
        else:
            logging.info("#$+ passed")

        return self.valid

    def bankid(self, pagename, bankname, delim="__"):
        return f"{pagename}{delim}{bankname}"

    # @logutil.loginfo(level='debug')
    def normalize(self, descriptor):
        """Auto-trims and -fills the descriptor
         - Removes any dangling bankids, assuming 'banks' keys as source of truth
         - Generates full bankid (pageid + bankname)
         - Handle most defaults here so they aren't scattered about
        """
        logging.info("#^Normalizing the input descriptor...")
        # TODO  Handle IDs here, perhaps with a utility class
        desc = copy.deepcopy(descriptor)

        desc["name"] = desc.get("name", "unspecified_name")
        desc["theme"] = desc.get("theme", "light")
        desc["data"] = desc.get("data", {})
        for data in desc["data"].values():
            # The default function call is assumed to be "load"
            if "call" not in data:
                data["call"] = "load"
            if "args" not in data:
                data["args"] = {}

        for pagename, page in desc["pages"].items():
            # Eliminate keys in connections that aren't in banks
            page["connections"] = dictutil.common_keys(
                page.get("connections", {}), page["banks"]
            )
            # Then rekey connections with both page and bank name
            page["connections"] = {
                self.bankid(pagename, from_bankname): {
                    self.bankid(pagename, to_bankname, delim="/")
                    for to_bankname in conn
                }
                for from_bankname, conn in page["connections"].items()
            }

            # Scan the layout for undefined banks and rekey
            new_layout = []
            for old_row in page.get("layout", []):
                new_row = []
                for bankname in old_row:
                    if bankname not in page["banks"]:
                        continue
                    new_item = {"bankid": self.bankid(pagename, bankname)}
                    new_row.append(new_item)
                new_layout.append(new_row)
            page["layout"] = new_layout

            # Scan the sidebar for undefined banks and rekey
            new_sidebar = []
            for bankname in page.get("sidebar", []):
                if bankname not in page["banks"]:
                    continue
                new_item = {"bankid": self.bankid(pagename, bankname)}
                new_item["title"] = butil.titlize(bankname)
                new_sidebar.append(new_item)
            page["sidebar"] = new_sidebar

            # Finally fix the banks
            new_banks = {}
            for bankname, bank in page["banks"].items():
                bank["args"] = bank.get("args", {})
                bank["args"]["gid"] = {"pageid": pagename, "bankid": bankname}
                new_banks[self.bankid(pagename, bankname)] = bank
            page["banks"] = new_banks
        logging.info("#$+ done")
        return desc

    def process_data(self, descriptor):
        logging.info("Loading the dataframes specified:")
        data = {}
        for dataid, entry in descriptor["data"].items():
            logging.info(f"#^  {dataid}...")
            try:
                data_module = importlib.import_module(entry["module"])
            except ImportError:
                logging.warning(f"\nFailed to load {entry['module']}")
                continue
            data[dataid] = getattr(data_module, entry["call"])(**entry["args"])
            data[dataid]["columns"] = list(data[dataid]["types"].keys())
            logging.info("#$+ done")
        return data

    def init_structure(self):
        logging.info("#^Generating initial context object...")
        # Outputs are a collection of fields provided by banks that are available
        # to be connected to other banks if the connections are provided
        # e.g. outputs + connections => connectors
        self.outputs = {}

        # Connectors define how the banks/components share information
        # They specify many-to-many <component_id>.<attribute> mappings
        # e.g. [axis_control.x_column, axis_control.y_column] -> graph.figure
        connectors = {
            "page": {
                "inputs": [("location", "pathname")],
                "outputs": [("page", "children")],
            }
        }

        # Callbacks are the boxes that the connectors connect. They take the inputs
        # and calculate the outputs, triggered when the inputs change.
        # TODO This needs generalization
        callback_code = """
            # Presumes only a simple url is passed in, probably enough for a template
            page_id = args[0].replace("/", "") if args[0] else "default"
            return page_index.get(page_id, list(page_index.values())[0])
            """

        callbacks = {
            "page": {"name": "update_page", "code": codeutil.format_code(callback_code)}
        }

        # This context contains the info fed into the Jinja template
        # The main job of the Bento class is to boil down the supplied description
        # into this context.
        self.theme_spec = style.BentoStyle(
            theme=self.desc.get("theme"), theme_dict=self.desc.get("theme_dict")
        ).spec
        self.context = {
            "name": self.desc["name"],
            "theme_spec": self.theme_spec,
            "appbar": self.desc.get("appbar", {}),
            "data": self.desc["data"],
            "pages": {},
            "banks": {},
            "connectors": connectors,
            "callbacks": callbacks,
        }
        logging.info("#$+ done")

    def create_page(self, pageid, page):
        # Prepares the definitions of the bank containers
        for bankid, bank in page["banks"].items():
            self.context["banks"][bankid] = self.build_bank(bank, page)

        self.outputs.update(self.bb.outputs)
        self.context["callbacks"].update(self.bb.callbacks)
        self.context["connectors"].update(self.bb.connectors)

        # Defines the layout of the page
        self.context["pages"][pageid] = grid.apply_grid(page)

    def connect_page(self, page):
        connections = page.get("connections", {})
        # Loop over all pairs of defined sources and sinks
        for source_bankid, sink_set in connections.items():
            for sink_regex in sink_set:
                for sink_cid, cb_def in self.context["callbacks"].items():
                    if not re.search(sink_regex, sink_cid):
                        continue
                    # Make sure we have a connector in place
                    if sink_cid not in self.context["connectors"]:
                        # The callback defines what component properties it provides
                        sink_vals = cb_def["provides"]
                        self.context["connectors"][sink_cid] = {
                            "outputs": [(sink_cid, val) for val in sink_vals],
                            "inputs": set(),
                        }
                    # Add any connected inputs
                    # TODO The bankid.replace needs to be improved
                    for cid, var in dictutil.extract(
                        source_bankid.replace("__", "/"), self.outputs, pop=False
                    ).items():
                        self.context["connectors"][sink_cid]["inputs"].add((cid, var))

    # @logutil.loginfo(level='debug')
    def build_bank(self, bank, page):
        # TODO Think through defaults on data, or perhaps just require a dataid
        args = {"dataid": "default"}
        # Get any page-level arguments
        args.update(dictutil.extract("dataid", page, pop=False))
        # Overwrite with specific args for bank
        args.update(bank["args"])
        item, sizing = getattr(self.bb, bank["type"])(**args)
        if "width" in bank:
            sizing["min"][1] = bank["width"]
            sizing["ideal"][1] = bank["width"]
        bank["sizing"] = sizing
        return item

    def write(self, app_output="bento_app.py", css_folder="assets"):
        if not self.valid:
            logging.warning("Descriptor never validated")
            return
        logging.info("Writing Dash application files:")
        self.write_template(self.context, self.template, app_output)

        pathlib.Path(css_folder).mkdir(parents=True, exist_ok=True)
        spec = style.BentoStyle(theme_dict=self.theme_spec).spec

        out_file = self.baseline_template.replace(".j2", "")
        self.write_template(spec, self.baseline_template, f"{css_folder}/{out_file}")

        out_file = self.theme_template.replace(".j2", "")
        self.write_template(spec, self.theme_template, f"{css_folder}/{out_file}")

    def write_template(self, context, template, output_location):
        env_args = {"trim_blocks": True, "lstrip_blocks": True}
        jenv = Environment(loader=PackageLoader(__name__), **env_args)
        template = jenv.get_template(template)
        str_output = template.render(context)

        # TODO Probably harden this against leaks
        if ".py" in output_location:
            mode = black.FileMode()
            try:
                str_output = black.format_file_contents(
                    str_output, fast=True, mode=mode
                )
            except Exception:
                logging.warning("Issue using Black to format file output")

        with open(output_location, "w") as fh:
            fh.write(str_output)

        logging.info(f"  wrote {output_location}")
